package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.util.TreeNode;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO.MetadataFile;
import org.eclipse.jgit.errors.InvalidPatternException;
import org.eclipse.jgit.ignore.FastIgnoreRule;
import org.eclipse.jgit.ignore.IMatcher;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;

/**
 * service to re-package a gitlab zip. Chop off leading root directory and
 * filter out unpublished entries.
 *
 * @author Michael Breu
 */
@Service
public class ZipRepackagingService {

    private static final Logger log = LoggerFactory.getLogger(ZipRepackagingService.class);

    public static final String PATH_SEPARATOR_STRING = Character.toString(FastIgnoreRule.PATH_SEPARATOR);

    /**
     * chops of the main folder, and brings every file one level up. Also
     * deleting all plain files in main folder.
     *
     * @param zipIn
     *            the zip to be repackaged
     * @param originalLocation
     *            the original location. For debug purposes only.
     * @param filterOut
     *            an array of file or foldernames that should be filtered away
     * @param exerciseFolderPath
     *            the path to the exercises
     * @return
     * @throws IOException
     */
    @SuppressWarnings({ "PMD.AvoidDeeplyNestedIfStmts", "PMD.DoNotUseThreads", "PMD.CognitiveComplexity", "PMD.UseTryWithResources" })
    protected InputStream rePackageGitLabProjectZip(
        ZipInputStream zipIn,
        String originalLocation,
        Optional<TreeNode<SearchResultDTO>> currentCollectionRoot,
        String exerciseFolderPath
    ) throws IOException {
        final PipedInputStream pis = new PipedInputStream(1024 * 256);
        final PipedOutputStream pos = new PipedOutputStream(pis);

        Runnable rePackage = () -> {
            try (ZipOutputStream zipOut = new ZipOutputStream(pos)) {
                final List<String> applicableFilters = currentCollectionRoot.isEmpty()
                    ? java.util.Collections.emptyList()
                    : getApplicableFilters(exerciseFolderPath, currentCollectionRoot.get());
                ZipEntry zipInEntry = zipIn.getNextEntry();
                String prefix = zipInEntry.isDirectory() ? zipInEntry.getName() : "";
                if (prefix.endsWith(PATH_SEPARATOR_STRING)) {
                    prefix = prefix.substring(0, prefix.length() - 1);
                }
                final String exercisePath = normalizeAbsolutePath(exerciseFolderPath);

                while (zipInEntry != null) {
                    String newName = filterAndNewName(zipInEntry, applicableFilters, exercisePath, prefix);
                    if (newName != null) {
                        if (newName.startsWith("/")) {
                            newName = newName.substring(1);
                        }
                        ZipEntry zipOutEntry = new ZipEntry(newName);
                        if (zipInEntry.isDirectory()) {
                            zipOut.putNextEntry(zipOutEntry);
                        } else {
                            zipOut.putNextEntry(zipOutEntry);
                            StreamUtils.copy(zipIn, zipOut);
                            zipOut.closeEntry();
                        }
                    }
                    zipInEntry = zipIn.getNextEntry();
                }
            } catch (IOException e) {
                log.error("Cannot rezip file from {}", originalLocation, e);
            }
        };

        new Thread(rePackage).start();
        return pis;
    }

    /**
     * returns the new name, if entry should be included.
     * returns null, if entry should be dropped.
     *
     * @param zipInEntry     the zipped Entry
     * @param collectionRoot the hierarchy of metadata for this project (from root)
     * @param gitlabPrefix   some (artificial) prefix exported from gitlab
     * @return
     */
    protected String filterAndNewName(ZipEntry zipInEntry, List<String> applicableFilters, String exerciseFolderPath, String gitlabPrefix) {
        String fileName = zipInEntry.getName();
        boolean isDirectory = zipInEntry.isDirectory();
        if (fileName.startsWith(gitlabPrefix)) {
            String rootProjectRelativeName = fileName.substring(gitlabPrefix.length());
            String fileWithSubPath = getWithoutPathPrefix(rootProjectRelativeName, exerciseFolderPath);
            // Fix for issue #415
            if (
                fileWithSubPath == null ||
                "/".equals(fileWithSubPath) ||
                exerciseFolderPath != null &&
                !exerciseFolderPath.endsWith("/") &&
                !fileWithSubPath.startsWith("/")
            ) {
                return null;
            }
            if (!ObjectUtils.isEmpty(fileWithSubPath) && !filterOut(rootProjectRelativeName, applicableFilters, isDirectory)) {
                return fileWithSubPath;
            }
        }
        return null;
    }

    public String getExerciseFolderPath(String exercisesPath) {
        if (exercisesPath == null) {
            return "";
        }
        int slashPos = exercisesPath.lastIndexOf('/');
        if (slashPos < 0) {
            return ""; // root metadata
        }
        return exercisesPath.substring(0, slashPos);
    }

    /**
     * returns the dirpath without trailing slash
     *
     * @param file
     * @return
     */
    private static String getDirPath(MetadataFile file) {
        String absFileName = file.getPath();
        int lastSlash = absFileName.lastIndexOf(FastIgnoreRule.PATH_SEPARATOR);
        if (lastSlash < 0 || StringUtils.isEmpty(absFileName)) {
            return Character.toString(FastIgnoreRule.PATH_SEPARATOR);
        }
        return normalizeAbsolutePath(absFileName.substring(0, lastSlash));
    }

    /**
     * shortens the fileName by the pathPrefix. Returns null, if fileName does
     * not start with path prefix.
     *
     * @param fileName
     *            the file name
     * @param pathPrefix
     *            the path prefix
     * @return
     */
    private String getWithoutPathPrefix(String fileName, String pathPrefix) {
        if (ObjectUtils.isEmpty(pathPrefix)) {
            return fileName;
        }
        if (fileName.startsWith(pathPrefix)) {
            if (PATH_SEPARATOR_STRING.equals(pathPrefix)) {
                return fileName;
            }
            if (fileName.length() == pathPrefix.length()) {
                return "/";
            } else {
                return fileName.substring(pathPrefix.length());
            }
        }
        return null;
    }

    protected static boolean filterOut(final String fullPath, final List<String> applicableFilters, boolean isDirectory) {
        for (String filter : applicableFilters) {
            if (filter == null) {
                log.warn("filter should never be null!");
                continue;
            }
            IMatcher matcher;
            try {
                if (filter.endsWith(Character.toString(FastIgnoreRule.PATH_SEPARATOR))) {
                    matcher = IMatcher.createPathMatcher(filter.substring(0, filter.length() - 1), true);
                } else {
                    matcher = IMatcher.createPathMatcher(filter, false);
                }
            } catch (InvalidPatternException e) {
                log.warn("Cannot parse pattern {}", filter);
                continue;
            }

            if (matcher.matches(fullPath, isDirectory, false)) {
                return true;
            }
        }
        return false;
    }

    /**
     * normalizes an absolute path. I.e. it prefixes it with an slash if none
     * provided.
     *
     * @param path
     * @return
     */
    private static String normalizeAbsolutePath(String path) {
        if (StringUtils.isEmpty(path)) {
            return PATH_SEPARATOR_STRING;
        }
        if (path.charAt(0) == FastIgnoreRule.PATH_SEPARATOR) {
            return path;
        }
        return PATH_SEPARATOR_STRING + path;
    }

    /**
     * gets all applicable filters for all nodes within current path
     *
     * @param exercisePath    the path to current exercise ( the item path is
     *                        absolute (i.e.
     *                        relativ to project root) and may or not start with a
     *                        slash)
     * @param currentTreeNode (path to exercise). The path to the meta data file
     *                        must be absolute relative to git project root. The exclusion pathes inside the current Treenode may be absolute (to current Exercise Path, or relative)
     * @return
     */

    @SuppressWarnings({ "PMD.CognitiveComplexity", "PMD.CyclomaticComplexity" })
    protected static List<String> getApplicableFilters(final String exercisePathF, TreeNode<SearchResultDTO> currentTreeNode) {
        if (exercisePathF == null) {
            return new ArrayList<>();
        }
        final String exercisePath = normalizeAbsolutePath(exercisePathF);
        // String relativeToRootExercisePath =
        // path.substring(rootExercisePath.length())

        final String dirPath = getDirPath(currentTreeNode.getData().getFile());
        final int dirPathLength = dirPath.length();

        // four possibilities
        // 1 + 2. currentTreeNode is a true ancestor or exactly the exercise
        // => collect all exceptions if potentially applicable to current path
        // 2. currentTreeNode is exactly the exercise
        // => collect all exceptions (and also concatenate with parent path) !Warning!
        // this could be done more accurate, if a more concrete specification is
        // developed.
        // 3. currentTreeNode is a true descendant
        // => collect all exceptions and prepend then by the parent paths
        // 4. currentTreeNode and current exercisePath are completely disjoint
        // => do nothing

        List<String> result = new ArrayList<>();

        if (
            StringUtils.isEmpty(dirPath) ||
            exercisePath.equals(dirPath) ||
            (
                exercisePath.length() > dirPathLength &&
                exercisePath.startsWith(dirPath) &&
                (exercisePath.charAt(dirPathLength) == FastIgnoreRule.PATH_SEPARATOR || PATH_SEPARATOR_STRING.equals(dirPath))
            )
        ) {
            // case 1 + 2: currentTreeNode is a true ancestor or same
            Arrays
                .stream(currentTreeNode.getData().getMetadata().getPublicVisibility().getExcept())
                .forEach(s -> {
                    if (s.length() > 1 && s.charAt(0) == FastIgnoreRule.PATH_SEPARATOR) {
                        if (PATH_SEPARATOR_STRING.equals(dirPath)) {
                            result.add(s);
                        } else {
                            result.add(dirPath + s);
                        }
                    } else {
                        if (PATH_SEPARATOR_STRING.equals(dirPath)) {
                            result.add("**" + FastIgnoreRule.PATH_SEPARATOR + s);
                        } else {
                            result.add(dirPath + FastIgnoreRule.PATH_SEPARATOR + "**" + FastIgnoreRule.PATH_SEPARATOR + s);
                        }
                    }
                });
        } else {
            if (
                dirPathLength > exercisePath.length() &&
                dirPath.startsWith(exercisePath) &&
                (dirPath.charAt(exercisePath.length()) == FastIgnoreRule.PATH_SEPARATOR || PATH_SEPARATOR_STRING.equals(exercisePath))
            ) {
                // case 3: this is a child
                Arrays
                    .stream(currentTreeNode.getData().getMetadata().getPublicVisibility().getExcept())
                    .forEach(s -> {
                        if (s.length() > 1 && s.charAt(0) == FastIgnoreRule.PATH_SEPARATOR || StringUtils.isEmpty(dirPath)) {
                            result.add(s);
                        } else {
                            result.add(dirPath + FastIgnoreRule.PATH_SEPARATOR + s);
                        }
                    });
            } else {
                log.debug("rootExercisePath {} is not a substring of current directory path {}, we can ignore it!", exercisePath, dirPath);
                return java.util.Collections.emptyList();
            }
        }
        //
        //            if (currentTreeNode.getData().getMetadata().getPublicVisibility().getExcept() != null) {
        //                if ("".equals(dirPath)) {
        //                    // if in root, just use the current filters
        //                    Arrays.stream(currentTreeNode.getData().getMetadata().getPublicVisibility().getExcept()).forEach(result::add);
        //                } else {
        //                    // if not in root, check prepending path
        //                    Arrays
        //                        .stream(currentTreeNode.getData().getMetadata().getPublicVisibility().getExcept())
        //                        .forEach(filterString -> {
        //                            if (filterString.startsWith(Character.toString(FastIgnoreRule.PATH_SEPARATOR))) {
        //                                result.add(FastIgnoreRule.PATH_SEPARATOR + dirPath + filterString);
        //                            } else if (!filterString.startsWith("**") && filterString.indexOf(FastIgnoreRule.PATH_SEPARATOR) >= 0) {
        //                                result.add(FastIgnoreRule.PATH_SEPARATOR + dirPath + FastIgnoreRule.PATH_SEPARATOR + filterString);
        //                            } else {
        //                                result.add(filterString);
        //                            }
        //                        });
        //                }
        //            }
        currentTreeNode.getChildren().forEach(childNode -> result.addAll(getApplicableFilters(exercisePath, childNode)));
        return result;
        //        } else {
        //            return java.util.Collections.emptyList();
        //        }
    }

    protected static List<String> getApplicableOfCurrentAsParent(String exercisePath, TreeNode<SearchResultDTO> currentTreeNode) {
        return java.util.Collections.emptyList();
    }
}
